﻿//==============================================================================
// File: NotifiableProperty.cs
// Created: 2010-05-10
// Author: Piotr Włodek
//==============================================================================
// This file is a part of MAMMOTH project.
// Copyright (C) 2010 AGH University of Science and Technology, Krakow.
// https://caribou.iisg.agh.edu.pl/trac/mammoth
//
// Mammoth is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Mammoth is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with WpfRcp. If not, see http://www.gnu.org/licenses/.
//==============================================================================

using System;
using System.Diagnostics;

namespace WpfRcp.PropertyModel.Core
{
    /// <summary>
    /// Property with the ability to notify its environment that the inner value has changed.
    /// </summary>
    [DebuggerDisplay("Name = {MetaProperty.Name}, Value = {Value}, IsValid = {IsValid}, IsSynchronizing = {IsSynchronizing}")]
    [Serializable]
    public class NotifiableProperty : Property, INotifyObjectStateChanged
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="NotifiableProperty"/>.
        /// </summary>
        /// <param name="property">Meta property representing the proerty itself.</param>
        public NotifiableProperty(MetaProperty property) : base(property)
        {
        }

        /// <summary>
        /// Gets or sets a value.
        /// </summary>
        public override object Value
        {
            set
            {
                if (Value is INotifyObjectStateChanged)
                {
                    var notifiableOldValue = (INotifyObjectStateChanged)Value;
                    notifiableOldValue.ObjectStateChanged -= OnInnerValueStateChanged;
                }

                base.Value = value;

                if (Value is INotifyObjectStateChanged)
                {
                    var notifiableNewValue = (INotifyObjectStateChanged)Value;
                    notifiableNewValue.ObjectStateChanged += OnInnerValueStateChanged;
                }
            }
        }

        private void OnInnerValueStateChanged(object sender, EventArgs e)
        {
            if (ObjectStateChanged != null)
                ObjectStateChanged(this, EventArgs.Empty);
        }

        #region Implementation of INotifyObjectStateChanged

        /// <summary>
        /// Event raised whenever object's state changes.
        /// </summary>
        public event EventHandler<EventArgs> ObjectStateChanged;

        #endregion

        #region Protected Overrides

        protected override void Dispose(bool disposing)
        {
            try
            {
                if (Value is INotifyObjectStateChanged)
                {
                    var notifiableValue = (INotifyObjectStateChanged)Value;
                    notifiableValue.ObjectStateChanged -= OnInnerValueStateChanged;
                }
            }
            finally
            {
                base.Dispose(disposing);
            }

        }

        #endregion
    }
}