﻿//==============================================================================
// File: PropertyScopeBase.cs
// Created: 2010-05-10
// Author: Piotr Włodek
//==============================================================================
// This file is a part of MAMMOTH project.
// Copyright (C) 2010 AGH University of Science and Technology, Krakow.
// https://caribou.iisg.agh.edu.pl/trac/mammoth
//
// Mammoth is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Mammoth is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with WpfRcp. If not, see http://www.gnu.org/licenses/.
//==============================================================================

using System;
using Res = WpfRcp.PropertyModel.Properties.Resources;

namespace WpfRcp.PropertyModel.Scope
{
    /// <summary>
    /// Represents a base class for all property scopes.
    /// </summary>
    [Serializable]
    public abstract class PropertyScopeBase : IPropertyScope, INotifyScopeChanged, IDisposable
    {
        #region Public Interface

        /// <summary>
        /// Gets value, if this scope is sealed.
        /// </summary>
        public bool IsSealed { get; private set; }

        /// <summary>
        /// Seals this scope, which disables further property collection modification.
        /// </summary>
        public void Seal()
        {
            if (Disposed)
                throw new ObjectDisposedException(Res.ExceptionObjectAlreadyDisposed);

            if (IsSealed)
                throw new PropertyScopeException(Res.ExceptionPropertyScopeSealed);

            IsSealed = true;
        }

        /// <summary>
        /// Method that returns all dynamic properties registered in the registry.
        /// </summary>
        /// <param name="container">The container for which <see cref="MetaPropertyCollection"/> is being returned.</param>
        /// <returns>Returns all dynamic properties registered in the registry for given <see cref="IPropertyContainer"/> container.</returns>
        /// <remarks>The <paramref name="container"/> parameter can be null.</remarks>
        public abstract MetaPropertyCollection GetDynamicProperties(IPropertyContainer container);

        #endregion

        #region Dispose Implementation

        private bool m_Disposed;

        ~PropertyScopeBase()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            lock (this)
            {
                if (m_Disposed == false)
                {
                    Dispose(true);
                    m_Disposed = true;

                    GC.SuppressFinalize(this);
                }
            }
        }

        protected bool Disposed
        {
            get
            {
                lock (this)
                {
                    return m_Disposed;
                }
            }
        }

        protected virtual void Dispose(bool disposing)
        {
        }

        #endregion

        #region INotifyScopeChanged Implementation

        /// <summary>
        /// Event raised whenever property scope changes.
        /// </summary>
        [field: NonSerialized]
        public event EventHandler<ScopeChangedEventArgs> ScopeChanged;

        protected virtual void OnScopeChanged(MetaProperty newProperty, MetaProperty oldProperty, Type registrationType)
        {
            if (ScopeChanged != null)
                ScopeChanged(this, new ScopeChangedEventArgs(newProperty, oldProperty, registrationType));
        }

        #endregion
    }
}