﻿//==============================================================================
// File: AboutViewPresentationModel.cs
// Created: 2010-05-28
// Author: Piotr Włodek, Arkadiusz Świerczek
//==============================================================================
// This file is a part of MAMMOTH project.
// Copyright (C) 2010 AGH University of Science and Technology, Krakow.
// https://caribou.iisg.agh.edu.pl/trac/mammoth
//
// Mammoth is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Mammoth is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with Mammoth. If not, see http://www.gnu.org/licenses/.
//==============================================================================

using System;
using System.Globalization;
using System.Reflection;
using WpfRcp.Core.Globals;
using WpfRcp.Core.Services;
using WpfRcp.Core.Views;
using WpfRcp.Core.Views.Presenters;
using Microsoft.Practices.Composite.Events;
using Microsoft.Practices.Composite.Presentation.Commands;
using Microsoft.Practices.Unity;

namespace WpfRcp.Views.Presenters
{
    public class AboutViewPresentationModel : PresenterBase<IAboutView>
    {
        public DelegateCommand<object> CloseCommand { get; private set; }
        public DelegateCommand<object> OpenModuleInfoCommand { get; private set; }

        private readonly IUnityContainer m_Container;
        private readonly IRcpSettings m_RcpSettings;
        private string m_VersionString;
        
        public AboutViewPresentationModel(IAboutView view, IShell shell, IUnityContainer container, IEventAggregator eventAggregator, IRcpSettings rcpSettings)
        {
            m_Container = container;
            m_RcpSettings = rcpSettings;

            CloseCommand = new DelegateCommand<object>(ExecuteCloseCommand);
            OpenModuleInfoCommand = new DelegateCommand<object>(ExecuteOpenModulesInfoCommand);

            View = view;
            View.PresentationModel = this;
            View.ViewOwner = shell;

            // Bind data
            DataBind();

            eventAggregator.GetEvent<LanguageChangedEvent>().Subscribe(LanguageChangedHandler);
        }

        public string VersionString
        {
            get { return m_VersionString; }
            private set
            {
                m_VersionString = value;
                OnPropertyChanged("VersionString");
            }
        }

        public IRcpSettings RcpSettings
        {
            get { return m_RcpSettings; }
        }

        public void DataBind()
        {
#if (DEBUG)
            VersionString =
                string.Format(
                    (IntPtr.Size == 4
                         ? Properties.Resources.VersionString32
                         : Properties.Resources.VersionString64) + " (Debug)",
                    Assembly.GetExecutingAssembly().GetName().Version);
#else
            if (IntPtr.Size == 4)
                VersionString = string.Format(Properties.Resources.VersionString32,
                    Assembly.GetExecutingAssembly().GetName().Version);
            else
                VersionString = string.Format(Properties.Resources.VersionString64,
                    Assembly.GetExecutingAssembly().GetName().Version);
#endif
        }

        private void LanguageChangedHandler(CultureInfo args)
        {
            DataBind();
        }

        private void ExecuteOpenModulesInfoCommand(object param)
        {
            var presenter = m_Container.Resolve<ModulesInfoPresentationModel>();
            presenter.Show(View);
        }

        private void ExecuteCloseCommand(object param)
        {
            View.Close();

            var presenter = m_Container.Resolve<ModulesInfoPresentationModel>();
            if (presenter.IsShown)
                presenter.Close();
        }
    }
}